# This file is part of Ubuntu Continuous Integration configuration framework.
#
# Copyright 2013, 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
"""Configuration stores contains key/value definitions.

They may define a single level of sections, each of which containing key/value
definitions.
"""

from __future__ import unicode_literals

import collections
import errno
import io
import warnings

from uciconfig import (
    errors,
    parsers,
)


class Section(object):
    """A section defines a dict of option name => value.

    This is merely a read-only dict which can add some knowledge about the
    options. It is *not* a python dict object though and doesn't try to mimic
    its API.
    """

    def __init__(self, section_id, options):
        self.id = section_id
        # We re-use the dict-like object received
        self.options = options

    def get(self, name, default=None, expand=True):
        return self.options.get(name, default)

    def iter_option_names(self):
        for k in self.options.iterkeys():
            yield k

    def __repr__(self):
        # Mostly for debugging purposes
        return "<stores.%s id=%s>" % (self.__class__.__name__, self.id)


_NewlyCreatedOption = object()
"""Was the option created during the MutableSection lifetime"""
_DeletedOption = object()
"""Was the option deleted during the MutableSection lifetime"""


class MutableSection(Section):
    """A section allowing changes and keeping track of the original values."""

    def __init__(self, section_id, options):
        super(MutableSection, self).__init__(section_id, options)
        self.reset_changes()

    def set(self, name, value):
        if name not in self.options:
            # This is a new option
            self.orig[name] = _NewlyCreatedOption
        elif name not in self.orig:
            self.orig[name] = self.get(name, None)
        self.options[name] = value

    def remove(self, name):
        if name not in self.orig and name in self.options:
            self.orig[name] = self.get(name, None)
        del self.options[name]

    def reset_changes(self):
        self.orig = {}

    def apply_changes(self, dirty, store):
        """Apply option value changes.

        ``self`` has been reloaded from the persistent storage. ``dirty``
        contains the changes made since the previous loading.

        :param dirty: the mutable section containing the changes.

        :param store: the store containing the section
        """
        for k, expected in dirty.orig.iteritems():
            actual = dirty.get(k, _DeletedOption)
            if actual is _DeletedOption:
                if k in self.options:
                    self.remove(k)
            else:
                self.set(k, actual)
        # No need to keep track of these changes
        self.reset_changes()


class FileStore(object):
    """A configuration store using a local file storage.

    The file content must be utf8 encoded.

    :cvar readonly_section_class: The class used to create read-only sections.

    :cvar mutable_section_class: The calss used to create mutable sections.
    """

    readonly_section_class = Section
    mutable_section_class = MutableSection

    def __init__(self, path):
        """A Store using a file on disk.

        :param path: The config file path.
        """
        super(FileStore, self).__init__()
        self.path = path
        self.sections = collections.OrderedDict()
        # Daughter classes can use a more specific id
        self.id = path
        # Which sections need to be saved (by section id). We use a dict here
        # so the dirty sections can be shared by multiple callers.
        self.dirty_sections = {}
        self._loaded = False

    def is_loaded(self):
        return self._loaded

    def unload(self):
        self._loaded = False
        self.dirty_sections = {}

    def _load_content(self):
        """Load the config file bytes.

        :return: Unicode string.
        """
        try:
            with io.open(self.path, encoding='utf8') as f:
                return f.read()
        except UnicodeDecodeError:
            raise errors.InvalidContent(self.path)
        except IOError as e:
            if e.errno == errno.EACCES:
                warnings.warn('Permission denied while trying to load'
                              ' configuration store {}.'
                              .format(self.external_url()))
                raise errors.PermissionDenied(self.path)
            elif e.errno in (errno.ENOENT, errno.ENOTDIR):
                raise errors.NoSuchFile(self.path)
            else:
                raise

    def _save_content(self, content):
        """Save the config file bytes.

        This should be provided by subclasses

        :param content: Config file unicode string to write
        """
        with io.open(self.path, 'w', encoding='utf8') as f:
            f.write(content)

    def load(self):
        """Load the store from the associated file."""
        if self.is_loaded():
            return
        content = self._load_content()
        self._load_from_string(content)

    def _load_from_string(self, text):
        """Create a config store from a string.

        :param text: A unicode string representing the file content.
        """
        if self.is_loaded():
            raise AssertionError('Already loaded')

        parser = parsers.Parser()
        tokens = parser.parse_config(self.path, text)
        for section in parser.make_sections(tokens):
            self.sections[section.name] = section
        self._loaded = True

    def _needs_saving(self):
        for s in self.dirty_sections.values():
            if s.orig:
                # At least one dirty section contains a modification
                return True
        return False

    # FIXME: Needs more testing -- vila 1014-01-13
    def apply_changes(self, dirty_sections):
        """Apply changes from dirty sections while checking for coherency.

        The Store content is discarded and reloaded from persistent storage to
        acquire up-to-date values.

        Dirty sections are MutableSection which kept track of the value they
        are expected to update.
        """
        # We need an up-to-date version from the persistent storage, unload the
        # store. The reload will occur when needed (triggered by the first
        # get_mutable_section() call below.
        self.unload()
        # Apply the changes from the preserved dirty sections
        for section_id, dirty in dirty_sections.iteritems():
            clean = self.get_mutable_section(section_id)
            clean.apply_changes(dirty, self)
        # Everything is clean now
        self.dirty_sections = {}

    # FIXME: Needs more testing -- vila 1014-01-13
    def save_changes(self):
        if not self.is_loaded():
            # Nothing to save
            return
        if not self._needs_saving():
            return
        # Preserve the current version
        dirty_sections = dict(self.dirty_sections.items())
        self.apply_changes(dirty_sections)
        # Save to the persistent storage
        self.save()

    def save(self):
        if not self.is_loaded():
            # Nothing to save
            return
        chunks = []
        for section in self.sections.values():
            chunks.append(section.serialize())
        self._save_content(''.join(chunks))

    def get_sections(self):
        """Get the section in the file order.

        :returns: An iterable of (store, section).
        """
        # We need a loaded store
        try:
            self.load()
        except (errors.NoSuchFile, errors.PermissionDenied):
            # If the file can't be read, there is no sections
            return
        for section_name in self.sections.keys():
            yield (self,
                   self.readonly_section_class(section_name,
                                               self.sections[section_name]))

    def get_mutable_section(self, section_id=None):
        # We need a loaded store
        try:
            self.load()
        except errors.NoSuchFile:
            # The file doesn't exist, let's pretend it was empty
            self._load_from_string('')
        if section_id in self.dirty_sections:
            # We already created a mutable section for this id
            return self.dirty_sections[section_id]
        section = self.sections.get(section_id, None)
        if section is None:
            section = parsers.Section(section_id)
            self.sections[section_id] = section
        mutable_section = self.mutable_section_class(section_id, section)
        # All mutable sections can become dirty
        self.dirty_sections[section_id] = mutable_section
        return mutable_section

    def quote(self, value):
        # We don't support quoting for now
        return value

    def unquote(self, value):
        # We don't support quoting for now
        return value

    def external_url(self):
        return self.path



# This file is part of Ubuntu Continuous Integration configuration framework.
#
# Copyright 2013, 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.
# Copyright (c) 2013 Canonical. See COPYING for details.

from ucitests.tests import test_styles

import uciconfig


class TestPep8(test_styles.TestPep8):

    packages = [uciconfig]


class TestPyflakes(test_styles.TestPyflakes):

    packages = [uciconfig]



# This file is part of Ubuntu Continuous Integration configuration framework.
#
# Copyright 2013, 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import unittest
import warnings


from uciconfig import (
    errors,
    options,
    registries,
)
from ucitests import (
    assertions,
    fixtures,
)


class TestOption(unittest.TestCase):

    def test_default_value(self):
        opt = options.Option('foo', 'foo help.', default='bar')
        self.assertEqual('bar', opt.get_default())

    def test_callable_default_value(self):
        def bar():
            return 'bar'
        opt = options.Option('foo', 'foo help.', default=bar)
        self.assertEqual('bar', opt.get_default())

    def test_mandatory_option(self):
        opt = options.Option('foo', 'Foo help.', default=options.MANDATORY)
        with self.assertRaises(errors.OptionMandatoryValueError) as cm:
            opt.get_default()
        self.assertEqual('foo must be set.', unicode(cm.exception))

    def test_default_value_from_env(self):
        opt = options.Option('foo', 'foo help.', default='bar',
                             default_from_env=['FOO'])
        fixtures.override_env(self, 'FOO', 'quux')
        # Env variable provides a default taking over the option one
        self.assertEqual('quux', opt.get_default())

    def test_first_default_value_from_env_wins(self):
        opt = options.Option('foo', 'foo help.', default='bar',
                             default_from_env=['NO_VALUE', 'FOO', 'BAZ'])
        fixtures.override_env(self, 'FOO', 'foo')
        fixtures.override_env(self, 'BAZ', 'baz')
        # The first env var set wins
        self.assertEqual('foo', opt.get_default())

    def test_not_supported_list_default_value(self):
        with self.assertRaises(AssertionError):
            options.Option('foo', 'foo help.', default=[1])

    def test_not_supported_object_default_value(self):
        with self.assertRaises(AssertionError):
            options.Option('foo', 'foo help.', default=object())

    def test_not_supported_callable_default_value_not_unicode(self):
        def bar_not_unicode():
            return b'bar'
        opt = options.Option('foo', 'foo help.', default=bar_not_unicode)
        with self.assertRaises(AssertionError):
            opt.get_default()

    def test_get_help_topic(self):
        opt = options.Option('foo', 'foo help.')
        self.assertEqual('foo', opt.get_help_topic())


class TestBoolFromStore(unittest.TestCase):

    def assertIsTrue(self, s, accepted_values=None):
        res = options.bool_from_store(s, accepted_values=accepted_values)
        self.assertEqual(True, res)

    def assertIsFalse(self, s, accepted_values=None):
        res = options.bool_from_store(s, accepted_values=accepted_values)
        self.assertEqual(False, res)

    def assertIsNone(self, s, accepted_values=None):
        res = options.bool_from_store(s, accepted_values=accepted_values)
        self.assertIs(None, res)

    def test_know_valid_values(self):
        self.assertIsTrue('true')
        self.assertIsFalse('false')
        self.assertIsTrue('1')
        self.assertIsFalse('0')
        self.assertIsTrue('on')
        self.assertIsFalse('off')
        self.assertIsTrue('yes')
        self.assertIsFalse('no')
        self.assertIsTrue('y')
        self.assertIsFalse('n')
        # Also try some case variations
        self.assertIsTrue('True')
        self.assertIsFalse('False')
        self.assertIsTrue('On')
        self.assertIsFalse('Off')
        self.assertIsTrue('ON')
        self.assertIsFalse('OFF')
        self.assertIsTrue('oN')
        self.assertIsFalse('oFf')

    def test_invalid_values(self):
        self.assertIsNone(None)
        self.assertIsNone('doubt')
        self.assertIsNone('frue')
        self.assertIsNone('talse')
        self.assertIsNone('42')

    def test_provide_accepted_values(self):
        av = dict(y=True, n=False, yes=True, no=False)
        self.assertIsTrue('y', av)
        self.assertIsTrue('Y', av)
        self.assertIsTrue('Yes', av)
        self.assertIsFalse('n', av)
        self.assertIsFalse('N', av)
        self.assertIsFalse('No', av)
        self.assertIsNone('1', av)
        self.assertIsNone('0', av)
        self.assertIsNone('on', av)
        self.assertIsNone('off', av)


def assertConverted(test, expected, opt, value):
    test.assertEqual(expected, opt.convert_from_unicode(None, value))


def assertWarns(test, opt, value):
    with warnings.catch_warnings(record=True) as wcm:
        test.assertEqual(None, opt.convert_from_unicode(None, value))
    assertions.assertLength(test, 1, wcm)
    test.assertEqual(
        'Value "{}" is not valid for "{}"'.format(value, opt.name),
        str(wcm[0].message))


def assertErrors(test, opt, value):
    with test.assertRaises(errors.OptionValueError):
        opt.convert_from_unicode(None, value)


def assertConvertInvalid(test, opt, invalid_value):
    opt.invalid = None
    test.assertEqual(None, opt.convert_from_unicode(None, invalid_value))
    opt.invalid = 'warning'
    assertWarns(test, opt, invalid_value)
    opt.invalid = 'error'
    assertErrors(test, opt, invalid_value)


class TestOptionWithBooleanConverter(unittest.TestCase):

    def get_option(self):
        return options.Option('foo', 'A boolean.',
                              from_unicode=options.bool_from_store)

    def test_convert_invalid(self):
        opt = self.get_option()
        # Not all strings are recognized as a boolean
        assertConvertInvalid(self, opt, 'invalid-boolean')
        # A list of strings is never recognized as a boolean
        assertConvertInvalid(self, opt, ['not', 'a', 'boolean'])

    def test_convert_valid(self):
        opt = self.get_option()
        assertConverted(self, True, opt, 'True')
        assertConverted(self, True, opt, '1')


class TestOptionWithIntegerConverter(unittest.TestCase):

    def get_option(self):
        return options.Option('foo', 'An integer.',
                              from_unicode=options.int_from_store)

    def test_convert_invalid(self):
        opt = self.get_option()
        # A string that is not recognized as an integer
        assertConvertInvalid(self, opt, 'forty-two')
        # A list of strings is never recognized as an integer
        assertConvertInvalid(self, opt, ['a', 'list'])

    def test_convert_valid(self):
        opt = self.get_option()
        assertConverted(self, 16, opt, '16')


class TestOptionWithSIUnitConverter(unittest.TestCase):

    def get_option(self):
        return options.Option('foo', 'An integer in SI units.',
                              from_unicode=options.int_SI_from_store)

    def test_convert_invalid(self):
        opt = self.get_option()
        assertConvertInvalid(self, opt, 'not-a-unit')
        assertConvertInvalid(self, opt, 'Gb')  # Forgot the int
        assertConvertInvalid(self, opt, '1b')  # Forgot the unit
        assertConvertInvalid(self, opt, '1GG')
        assertConvertInvalid(self, opt, '1Mbb')
        assertConvertInvalid(self, opt, '1MM')

    def test_convert_valid(self):
        opt = self.get_option()
        assertConverted(self, int(5e3), opt, '5kb')
        assertConverted(self, int(5e6), opt, '5M')
        assertConverted(self, int(5e6), opt, '5MB')
        assertConverted(self, int(5e9), opt, '5g')
        assertConverted(self, int(5e9), opt, '5gB')
        assertConverted(self, 100, opt, '100')


class TestRegistryOption(unittest.TestCase):

    def get_option(self, registry, default=None):
        return options.RegistryOption('fooreg', 'A registry option.', registry,
                                      default=default)

    def test_convert_invalid(self):
        registry = registries.Registry()
        opt = self.get_option(registry)
        assertConvertInvalid(self, opt, [1])
        assertConvertInvalid(self, opt, 'notregistered')

    def test_convert_valid(self):
        registry = registries.Registry()
        registry.register('someval', 1234, 'Some help.')
        opt = self.get_option(registry)
        assertConverted(self, 1234, opt, 'someval')
        assertConverted(self, None, opt, None)

    def test_help(self):
        registry = registries.Registry()
        registry.register('someval',  1234, 'Some option.')
        registry.register('dunno', 1234, 'Some other option.')
        opt = self.get_option(registry)
        self.assertEqual('''A registry option.

The following values are supported:
 dunno - Some other option.
 someval - Some option.
''',
                         opt.help)

    def test_default_value(self):
        registry = registries.Registry()
        opt = self.get_option(registry, default='someval')
        self.assertEqual('someval', opt.get_default())


class TestOptionRegistry(unittest.TestCase):

    def setUp(self):
        super(TestOptionRegistry, self).setUp()
        # Always start with an empty registry
        fixtures.patch(self, options, 'option_registry',
                       options.OptionRegistry())
        self.registry = options.option_registry

    def test_register(self):
        opt = options.Option('foo', 'foo help.')
        self.registry.register(opt)
        self.assertIs(opt, self.registry.get('foo'))

    def test_registered_help(self):
        opt = options.Option('foo', 'A simple option')
        self.registry.register(opt)
        self.assertEquals('A simple option', self.registry.get_help('foo'))


class TestListOption(unittest.TestCase):

    def get_option(self):
        return options.ListOption('foo', 'A list.')

    def test_convert_invalid(self):
        opt = self.get_option()
        # We don't even try to convert a list into a list, we only expect
        # strings
        assertConvertInvalid(self, opt, [1])
        # No string is invalid as all forms can be converted to a list

    def test_convert_valid(self):
        opt = self.get_option()
        # An empty string is an empty list
        assertConverted(self, [], opt, '')
        # A boolean
        assertConverted(self, ['True'], opt, 'True')
        # An integer
        assertConverted(self, ['42'], opt, '42')
        # A single string
        assertConverted(self, ['bar'], opt, 'bar')



# This file is part of Ubuntu Continuous Integration configuration framework.
#
# Copyright 2013, 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest


from uciconfig import (
    errors,
    parsers,
)


# Some aliases to make tests shorter
SD = parsers.SectionDefinition
OD = parsers.OptionDefinition


def assertTokens(test, expected, to_parse):
    test.assertEqual(expected,
                     test.parser.parse_config('<string>', to_parse))


class TestParseConfigEdgeCases(unittest.TestCase):

    def setUp(self):
        super(TestParseConfigEdgeCases, self).setUp()
        self.parser = parsers.Parser()

    def test_empty(self):
        assertTokens(self, [], '')

    def test_almost_empty(self):
        # Empty lines or lines containing only spaces are ignored
        assertTokens(self, [], ' \n   \n\n   \n   ')

    def test_invalid(self):
        with self.assertRaises(errors.InvalidSyntax):
            self.parser.parse_config('<string>', 'whatever')


class TestParseKeyValues(unittest.TestCase):

    def setUp(self):
        super(TestParseKeyValues, self).setUp()
        self.parser = parsers.Parser()

    def test_single_key_value(self):
        assertTokens(self, [SD(None), OD('a', 'b')], 'a=b')

    def test_single_key_value_stripping_spaces(self):
        assertTokens(self, [SD(None), OD('a', 'b')], ' a = b ')

    def test_empty_value(self):
        assertTokens(self, [SD(None), OD('a', '')], 'a=')

    def test_pre_comment(self):
        assertTokens(self, [SD(None), OD('a', 'b', '# Define a\n')],
                     '# Define a\na=b')

    def test_post_comment(self):
        # The comment collect even the preceding spaces as they are not part of
        # the value
        assertTokens(self, [SD(None), OD('a', 'b', None, ' # b is good\n')],
                     'a=b # b is good\n')

    def test_several_options(self):
        assertTokens(self, [SD(None),
                            OD('a', 'b'),
                            OD('c', 'd', '# c now\n', ' # c is better\n')],
                     'a=b\n# c now\nc=d # c is better\n')


class TestParseSections(unittest.TestCase):

    def setUp(self):
        super(TestParseSections, self).setUp()
        self.parser = parsers.Parser()

    def test_empty_section(self):
        assertTokens(self, [SD('s')], '[s]')

    def test_empty_section_name(self):
        with self.assertRaises(errors.SectionEmptyName):
            self.parser.parse_config('<string>', '[]')

    def test_pre_comment(self):
        assertTokens(self, [SD('s', '  # Empty section\n')],
                     '  # Empty section\n[s]')

    def test_post_comment(self):
        assertTokens(self, [SD('s', None, ' # It is empty\n')],
                     '[s] # It is empty\n')


class TestSerializeSection(unittest.TestCase):

    def test_serialize_empty(self):
        sect = parsers.Section(None)
        self.assertEqual('', sect.serialize())

    def test_single_option(self):
        sect = parsers.Section(None)
        sect.add('foo', 'bar', '# This is foo\n', ' # Set to bar\n')
        self.assertEqual('''# This is foo
foo = bar # Set to bar
''',
                         sect.serialize())

    def test_single_option_in_named_section(self):
        sect = parsers.Section('sect', '# This is sect\n # A simple section\n',
                               ' # Just saying\n')
        sect.add('foo', 'bar', '# This is foo\n', ' # Set to bar\n')
        self.assertEqual('''# This is sect
 # A simple section
[sect] # Just saying
# This is foo
foo = bar # Set to bar
''',
                         sect.serialize())

    def test_several_options(self):
        sect = parsers.Section(None)
        sect.add('foo', 'bar', '# This is foo\n', ' # Set to bar\n')
        sect.add('baz', 'qux', '# This is baz\n', ' # Set to qux\n')
        self.assertEqual('''# This is foo
foo = bar # Set to bar
# This is baz
baz = qux # Set to qux
''',
                         sect.serialize())


class TestModifiedSectionSerialization(unittest.TestCase):

    def setUp(self):
        super(TestModifiedSectionSerialization, self).setUp()
        sect = parsers.Section(None)
        sect.add('b', '1', '# Define b\n', ' # set to 1\n')
        sect.add('a', '2', '# Define a\n', ' # set to 2\n')
        sect.add('z', '3', '# Define z\n', ' # set to 3\n')
        self.section = sect

    def test_added_option(self):
        self.section.add('foo', 'bar')
        self.assertEqual('''# Define b
b = 1 # set to 1
# Define a
a = 2 # set to 2
# Define z
z = 3 # set to 3
foo = bar
''',
                         self.section.serialize())

    def test_modified_option(self):
        self.section['a'] = '42'
        self.assertEqual('''# Define b
b = 1 # set to 1
# Define a
a = 42 # set to 2
# Define z
z = 3 # set to 3
''',
                         self.section.serialize())

    def test_deleted_option(self):
        del self.section['a']
        self.assertEqual('''# Define b
b = 1 # set to 1
# Define z
z = 3 # set to 3
''',
                         self.section.serialize())


class TestMakeSections(unittest.TestCase):

    def setUp(self):
        super(TestMakeSections, self).setUp()
        self.parser = parsers.Parser()

    def test_empty(self):
        self.assertEqual([], list(self.parser.make_sections([])))

    def test_single_section(self):
        defs = [SD(None), OD('a', 'b', None, ' # b is good\n')]
        sections = list(self.parser.make_sections(defs))
        self.assertEqual(1, len(sections))
        section = sections[0]
        self.assertEqual('b', section['a'])

    def test_several_sections(self):
        defs = [SD('s1'), OD('a', 'b', None, ' # b is good\n'),
                SD('s2'), OD('c', 'd'), OD('foo', 'bar')]
        sections = list(self.parser.make_sections(defs))
        self.assertEqual(2, len(sections))
        s1, s2 = sections
        self.assertEqual('b', s1['a'])
        self.assertEqual('d', s2['c'])
        self.assertEqual('bar', s2['foo'])



# This file is part of Ubuntu Continuous Integration configuration framework.
#
# Copyright 2013, 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import os
import re
import warnings


from uciconfig import (
    errors,
    registries,
)


class Option(object):
    """An option definition.

    The option *values* are stored in config files and found in sections.

    Here we define various properties about the option itself, its default
    value, how to convert it from stores, what to do when invalid values are
    encoutered, in which config files it can be stored.
    """

    def __init__(self, name, help_string, default=None,
                 default_from_env=None, override_from_env=None,
                 from_unicode=None, invalid=None, unquote=True):
        """Build an option definition.

        :param name: The name used to refer to the option.

        :param help_string: A doc string to explain the option to the user.

        :param default: The default value to use when none exist in the config
            stores. This is either a string that ``from_unicode`` will convert
            into the proper type, a callable returning a unicode string so that
            ``from_unicode`` can be used on the return value, or a python
            object that can be stringified (so only the empty list is supported
            for example). If an option must be set, the special MANDATORY
            (defined in this module) value can be used, an error is then raised
            if an attempt is made to access the option value.

        :param default_from_env: A list of environment variables which can
           provide a default value. 'default' will be used only if none of the
           variables specified here are set in the environment.

        :param override_from_env: A list of environment variables which can
           override any configuration setting.

        :param from_unicode: A callable to convert the unicode string
            representing the option value in a store or its default value.

        :param invalid: The action to be taken when an invalid value is
            encountered in a store. This is called only when from_unicode is
            invoked to convert a string and returns None or raise ValueError or
            TypeError. Accepted values are: None (ignore invalid values),
            'warning' (emit a warning), 'error' (emit an error message and
            terminates).

        :param unquote: Should the unicode value be unquoted before conversion.
           This should be used only when the store providing the values cannot
           safely unquote them. It is provided so daughter classes can handle
           the quoting themselves.
        """
        if override_from_env is None:
            override_from_env = []
        if default_from_env is None:
            default_from_env = []
        self.name = name
        self._help = help_string
        self.override_from_env = override_from_env
        # Convert the default value to a unicode string so all values are
        # strings internally before conversion (via from_unicode) is attempted.
        if default is None:
            self.default = None
        elif default is MANDATORY:
            # Special case, no further check needed
            self.default = default
        elif isinstance(default, list):
            # Only the empty list is supported
            if default:
                raise AssertionError(
                    'Only empty lists are supported as default values')
            self.default = ','
        elif isinstance(default, (unicode, bool, int, float)):
            # Rely on python to convert strings, booleans, floats and integers
            self.default = '{}'.format(default)
        elif callable(default):
            self.default = default
        else:
            # other python objects are not expected
            raise AssertionError(
                '{!r} is not supported as a default value'.format(default))
        self.default_from_env = default_from_env
        self.from_unicode = from_unicode
        self.unquote = unquote
        if invalid and invalid not in ('warning', 'error'):
            raise AssertionError(
                "{} not supported for 'invalid'".format(invalid))
        self.invalid = invalid

    @property
    def help(self):
        return self._help

    def convert_from_unicode(self, store, unicode_value):
        if self.unquote and store is not None and unicode_value is not None:
            unicode_value = store.unquote(unicode_value)
        if self.from_unicode is None or unicode_value is None:
            # Don't convert or nothing to convert
            return unicode_value
        try:
            converted = self.from_unicode(unicode_value)
        except (ValueError, TypeError):
            # Invalid values are ignored
            converted = None
        if converted is None and self.invalid is not None:
            # The conversion failed
            if self.invalid == 'warning':
                warnings.warn('Value "{}" is not valid'
                              ' for "{}"'.format(unicode_value, self.name))
            elif self.invalid == 'error':
                raise errors.OptionValueError(self.name, unicode_value)
        return converted

    def get_override(self):
        value = None
        for var in self.override_from_env:
            try:
                # If the env variable is defined, its value takes precedence
                value = os.environ[var]
                break
            except KeyError:
                continue
        return value

    def get_default(self):
        value = None
        for var in self.default_from_env:
            try:
                # If the env variable is defined, its value is the default one
                value = os.environ[var]
                break
            except KeyError:
                continue
        if value is None:
            # Otherwise, fallback to the value defined at registration
            if callable(self.default):
                value = self.default()
                if not isinstance(value, unicode):
                    raise AssertionError(
                        "Callable default value for '%s' should be unicode"
                        % (self.name))
            else:
                value = self.default
        if value is MANDATORY:
            raise errors.OptionMandatoryValueError(self.name)
        return value

    def get_help_topic(self):
        return self.name


# A special default value to indicate that an option must be set.
MANDATORY = object()


# Predefined converters to convert values from store into types other than
# unicode

_valid_boolean_strings = dict(yes=True, no=False,
                              y=True, n=False,
                              on=True, off=False,
                              true=True, false=False)
_valid_boolean_strings['1'] = True
_valid_boolean_strings['0'] = False


def bool_from_store(s, accepted_values=None):
    """Returns a boolean if the string can be interpreted as such.

    Interpret case insensitive strings as booleans. The default values
    includes: 'yes', 'no, 'y', 'n', 'true', 'false', '0', '1', 'on',
    'off'. Alternative values can be provided with the 'accepted_values'
    parameter.

    :param s: A string that should be interpreted as a boolean. It should be of
        type string or unicode.

    :param accepted_values: An optional dict with accepted strings as keys and
        True/False as values. The strings will be tested against a lowered
        version of 's'.

    :return: True or False for accepted strings, None otherwise.
    """
    if accepted_values is None:
        accepted_values = _valid_boolean_strings
    val = None
    if type(s) is unicode:
        try:
            val = accepted_values[s.lower()]
        except KeyError:
            pass
    return val


def int_from_store(string):
    return int(string)


_unit_suffixes = dict(K=10 ** 3, M=10 ** 6, G=10 ** 9)
_unit_re = re.compile("^(\d+)(([" + ''.join(_unit_suffixes) + "])b?)?$",
                      re.IGNORECASE)


def int_SI_from_store(string):
    """Convert a human readable size in SI units, e.g 10MB into an integer.

    Accepted suffixes are K,M,G. It is case-insensitive and may be followed
    by a trailing b (i.e. Kb, MB). This is intended to be practical and not
    pedantic.

    :param string: The option value as a unicode string.

    :return Integer, expanded to its base-10 value if a proper SI unit is
        found, None otherwise.
    """
    m = _unit_re.match(string)
    val = None
    if m is not None:
        val, _, unit = m.groups()
        val = int(val)
        if unit:
            try:
                coeff = _unit_suffixes[unit.upper()]
            except KeyError:
                raise ValueError('{0} is not an SI unit.'.format(unit))
            val *= coeff
    return val


def float_from_store(string):
    return float(string)


class ListOption(Option):

    def __init__(self, name, help_string, default=None,
                 default_from_env=None, override_from_env=None,
                 invalid=None):
        """A list Option definition.

        This overrides the base class so the conversion from a unicode string
        can take quoting (NIY) into account.
        """
        super(ListOption, self).__init__(name, help_string, default=default,
                                         default_from_env=default_from_env,
                                         override_from_env=override_from_env,
                                         from_unicode=self.from_unicode,
                                         invalid=invalid, unquote=False)

    def from_unicode(self, string):
        if not isinstance(string, unicode):
            raise TypeError
        return [s.strip() for s in string.split(',') if s]


class RegistryOption(Option):
    """Option for a choice from a registry."""

    def __init__(self, name, help_string, registry,
                 default=None, default_from_env=None,
                 invalid=None):
        """A registry based Option definition.

        This overrides the base class so the conversion from a unicode string
        can take quoting (NIY) into account.
        """
        super(RegistryOption, self).__init__(
            name, help_string, default=default,
            default_from_env=default_from_env,
            from_unicode=self.from_unicode,
            invalid=invalid, unquote=False)
        self.registry = registry

    def from_unicode(self, string):
        if not isinstance(string, unicode):
            raise TypeError
        try:
            return self.registry.get(string)
        except KeyError:
            raise ValueError(
                'Invalid value {} for {}.'
                'See help for a list of possible values.'.format(string,
                                                                 self.name))

    @property
    def help(self):
        ret = [self._help, "\n\nThe following values are supported:\n"]
        for key in sorted(self.registry.keys()):
            ret.append(" %s - %s\n" % (key, self.registry.get_help(key)))
        return "".join(ret)


class OptionRegistry(registries.Registry):
    """Register config options by their name.

    This overrides ``registries.Registry`` to simplify registration by
    acquiring some information from the option object itself.
    """

    def register(self, option):
        """Register a new option under its name.

        :param option: The option to register. Its name is used as the key.
        """
        super(OptionRegistry, self).register(option.name, option, option.help)


option_registry = OptionRegistry()



# This file is part of Ubuntu Continuous Integration configuration framework.
#
# Copyright 2013, 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import unittest
import warnings


from uciconfig import (
    errors,
    stores,
    tests,
)
from ucitests import (
    assertions,
    fixtures,
)


class TestSection(unittest.TestCase):

    def test_get_a_value(self):
        a_dict = dict(foo='bar')
        section = stores.Section('myID', a_dict)
        self.assertEquals('bar', section.get('foo'))

    def test_get_unknown_option(self):
        a_dict = dict()
        section = stores.Section(None, a_dict)
        self.assertEquals('out of thin air',
                          section.get('foo', 'out of thin air'))

    def test_options_is_shared(self):
        a_dict = dict()
        section = stores.Section(None, a_dict)
        self.assertIs(a_dict, section.options)


class TestMutableSection(unittest.TestCase):

    def get_section(self, opts):
        return stores.MutableSection('myID', opts)

    def test_set(self):
        a_dict = dict(foo='bar')
        section = self.get_section(a_dict)
        section.set('foo', 'new_value')
        self.assertEquals('new_value', section.get('foo'))
        # The change appears in the shared section
        self.assertEquals('new_value', a_dict.get('foo'))
        # We keep track of the change
        self.assertTrue('foo' in section.orig)
        self.assertEquals('bar', section.orig.get('foo'))

    def test_set_preserve_original_once(self):
        a_dict = dict(foo='bar')
        section = self.get_section(a_dict)
        section.set('foo', 'first_value')
        section.set('foo', 'second_value')
        # We keep track of the original value
        self.assertTrue('foo' in section.orig)
        self.assertEquals('bar', section.orig.get('foo'))

    def test_remove(self):
        a_dict = dict(foo='bar')
        section = self.get_section(a_dict)
        section.remove('foo')
        # We get None for unknown options via the default value
        self.assertEquals(None, section.get('foo'))
        # Or we just get the default value
        self.assertEquals('unknown', section.get('foo', 'unknown'))
        self.assertFalse('foo' in section.options)
        # We keep track of the deletion
        self.assertTrue('foo' in section.orig)
        self.assertEquals('bar', section.orig.get('foo'))

    def test_remove_new_option(self):
        a_dict = dict()
        section = self.get_section(a_dict)
        section.set('foo', 'bar')
        section.remove('foo')
        self.assertFalse('foo' in section.options)
        # The option didn't exist initially so it we need to keep track of it
        # with a special value
        self.assertTrue('foo' in section.orig)
        self.assertEquals(stores._NewlyCreatedOption, section.orig['foo'])


class TestFileStore(unittest.TestCase):

    def get_store(self):
        return stores.FileStore('foo.conf')

    def test_id(self):
        store = self.get_store()
        self.assertIsNot(None, store.id)

    def test_loading_unknown_file_fails(self):
        store = self.get_store()
        with self.assertRaises(errors.NoSuchFile):
            store.load()

    def test_invalid_content(self):
        store = self.get_store()
        self.assertFalse(store.is_loaded())
        with self.assertRaises(errors.InvalidSyntax) as cm:
            store._load_from_string('this is invalid !')
        self.assertEqual('foo.conf(1): Not a section nor an option.',
                         repr(cm.exception))
        # And the load failed
        self.assertFalse(store.is_loaded())


class TestReadOnlyFileStore(unittest.TestCase):

    def get_store(self):
        return stores.FileStore('foo.conf')

    def test_building_delays_load(self):
        store = self.get_store()
        self.assertEquals(False, store.is_loaded())
        store._load_from_string('')
        self.assertEquals(True, store.is_loaded())

    def test_get_no_sections_for_empty(self):
        store = self.get_store()
        store._load_from_string('')
        self.assertEquals([], list(store.get_sections()))

    def test_get_default_section(self):
        store = self.get_store()
        store._load_from_string('foo=bar')
        sections = list(store.get_sections())
        assertions.assertLength(self, 1, sections)
        tests.assertSectionContent(self, (None, {'foo': 'bar'}), sections[0])

    def test_get_named_section(self):
        store = self.get_store()
        store._load_from_string('[baz]\nfoo=bar')
        sections = list(store.get_sections())
        assertions.assertLength(self, 1, sections)
        tests.assertSectionContent(self, ('baz', {'foo': 'bar'}), sections[0])

    def test_load_from_string_fails_for_non_empty_store(self):
        store = self.get_store()
        store._load_from_string('foo=bar')
        self.assertRaises(AssertionError, store._load_from_string, 'bar=baz')


class TestFileStoreContent(unittest.TestCase):
    """Simulate loading a config store with content of various encodings.

    All files produced have an utf8 content.

    Users may modify them manually and end up with a file that can't be
    loaded. We need to issue proper error messages in this case.
    """

    invalid_utf8_char = b'\xff'

    def setUp(self):
        super(TestFileStoreContent, self).setUp()
        fixtures.set_uniq_cwd(self)

    def test_load_utf8(self):
        """Ensure we can load an utf8-encoded file."""
        # Store the raw content in the config file
        with open('foo.conf', 'wb') as f:
            f.write('user=b\N{Euro Sign}ar'.encode('utf8'))
        store = stores.FileStore('foo.conf')
        store.load()
        sections = list(store.get_sections())
        assertions.assertLength(self, 1, sections)
        s, section = sections[0]
        self.assertIs(store, s)
        self.assertEqual('b\N{Euro Sign}ar', section.get('user'))

    def test_load_non_ascii(self):
        """Ensure we display a proper error on non-ascii, non utf-8 content."""
        with open('foo.conf', 'wb') as f:
            f.write(b'user=foo\n#' + self.invalid_utf8_char + b'\n')
        store = stores.FileStore('foo.conf')
        with self.assertRaises(errors.InvalidContent):
            store.load()

    def test_load_erroneous_content(self):
        """Ensure we display a proper error on content that can't be parsed."""
        with open('foo.conf', 'w') as f:
            f.write('[open_section\n')
        store = stores.FileStore('foo.conf')
        with self.assertRaises(errors.InvalidSyntax):
            store.load()

    def test_load_permission_denied(self):
        """Ensure we get warned when trying to load an inaccessible file."""
        with open('foo.conf', 'w') as f:
            f.write('')
        os.chmod('foo.conf', 0o000)
        self.addCleanup(os.chmod, 'foo.conf', 0o644)
        store = stores.FileStore('foo.conf')
        with warnings.catch_warnings(record=True) as wcm:
            with self.assertRaises(errors.PermissionDenied):
                store.load()
        self.assertEqual('Permission denied while trying to load'
                         ' configuration store foo.conf.',
                         str(wcm[0].message))


class TestMutableStore(unittest.TestCase):

    def setUp(self):
        super(TestMutableStore, self).setUp()
        fixtures.set_uniq_cwd(self)

    def get_store(self):
        return stores.FileStore('foo.conf')

    def has_store(self, store):
        return os.access(store.path, os.F_OK)

    def test_save_empty_creates_no_file(self):
        store = self.get_store()
        store.save()
        self.assertEquals(False, self.has_store(store))

    def test_mutable_section_shared(self):
        store = self.get_store()
        store._load_from_string('foo=bar\n')
        section1 = store.get_mutable_section(None)
        section2 = store.get_mutable_section(None)
        # If we get different sections, different callers won't share the
        # modification
        self.assertIs(section1, section2)

    def test_save_emptied_succeeds(self):
        store = self.get_store()
        store._load_from_string('foo=bar\n')
        section = store.get_mutable_section(None)
        section.remove('foo')
        store.save()
        self.assertEquals(True, self.has_store(store))
        modified_store = self.get_store()
        sections = list(modified_store.get_sections())
        assertions.assertLength(self, 0, sections)

    def test_save_with_content_succeeds(self):
        store = self.get_store()
        store._load_from_string('foo=bar\n')
        self.assertEquals(False, self.has_store(store))
        store.save()
        self.assertEquals(True, self.has_store(store))
        modified_store = self.get_store()
        sections = list(modified_store.get_sections())
        assertions.assertLength(self, 1, sections)
        tests.assertSectionContent(self, (None, {'foo': 'bar'}), sections[0])

    def test_set_option_in_empty_store(self):
        store = self.get_store()
        section = store.get_mutable_section(None)
        section.set('foo', 'bar')
        store.save()
        modified_store = self.get_store()
        sections = list(modified_store.get_sections())
        assertions.assertLength(self, 1, sections)
        tests.assertSectionContent(self, (None, {'foo': 'bar'}), sections[0])

    def test_set_option_in_default_section(self):
        store = self.get_store()
        section = store.get_mutable_section(None)
        section.set('foo', 'bar')
        store.save()
        modified_store = self.get_store()
        sections = list(modified_store.get_sections())
        assertions.assertLength(self, 1, sections)
        tests.assertSectionContent(self, (None, {'foo': 'bar'}), sections[0])

    def test_set_option_in_named_section(self):
        store = self.get_store()
        store._load_from_string('')
        section = store.get_mutable_section('baz')
        section.set('foo', 'bar')
        store.save()
        modified_store = self.get_store()
        sections = list(modified_store.get_sections())
        assertions.assertLength(self, 1, sections)


class TestRoundTripping(unittest.TestCase):

    def setUp(self):
        super(TestRoundTripping, self).setUp()
        fixtures.set_uniq_cwd(self)
        self.store = stores.FileStore('foo.conf')

    def assertRoundTrip(self, content):
        self.store._load_from_string(content)
        self.store.save()
        with open(self.store.path) as f:
            new_content = f.read()
        assertions.assertMultiLineAlmostEqual(self, content, new_content)

    def test_raw(self):
        self.assertRoundTrip('''\
a = b
[s1]
c = d
[s2]
e = x
''')

    def test_with_comments(self):
        self.assertRoundTrip('''\
# Initial comment
a = b # And a comment
# section comment
[src] # another section comment
c = d
''')



# This file is part of Ubuntu Continuous Integration configuration framework.
#
# Copyright 2013 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.


def assertSectionContent(test, expected, actual):
    """Assert that some options have the proper values in a section.

    :param expected: A (name, options) tuple where 'name' is a section name and
        'options' a dict of the expection key/value pairs for the options.

    :param actual: A (store, section) tuple where 'store' is unused here but
       make the assertion easier to use and 'section' contains the options to
       check.
    """
    expected_name, expected_options = expected
    _, section = actual
    test.assertEquals(expected_name, section.id)
    test.assertEquals(expected_options,
                      dict([(k, section.get(k))
                            for k in expected_options.keys()]))



# This file is part of Ubuntu Continuous Integration configuration framework.
#
# Copyright 2013 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import unittest


from uciconfig import registries


class TestRegistry(unittest.TestCase):

    def setUp(self):
        super(TestRegistry, self).setUp()
        reg = registries.Registry()
        reg.register('one', 1, 'one help.')
        reg.register('two', 2, 'two help.')
        reg.register('four', 4, 'four help.')

        def five_help(reg, key):
            return 'five help.'
        reg.register('five', 5, five_help)

        self.reg = reg

    def test_get_unknown_key(self):
        with self.assertRaises(KeyError):
            self.reg.get('three')

    def test_get_existing_key(self):
        self.assertEqual(2, self.reg.get('two'))

    def test_get_help_unknown_key(self):
        with self.assertRaises(KeyError):
            self.reg.get_help('three')

    def test_get_help_existing_key(self):
        self.assertEqual('two help.', self.reg.get_help('two'))

    def test_get_callable_help_existing_key(self):
        self.assertEqual('five help.', self.reg.get_help('five'))

    def test_register_unknown_key(self):
        self.reg.register('forty-two', 42)
        self.assertEqual(42, self.reg.get('forty-two'))

    def test_register_existing_key(self):
        with self.assertRaises(KeyError):
            self.reg.register('one', 2)

    def test_remove_unknown_key(self):
        with self.assertRaises(KeyError):
            self.reg.remove('three')

    def test_remove_existing_key(self):
        self.reg.remove('one')
        with self.assertRaises(KeyError):
            self.reg.get('one')

    def test_keys(self):
        self.assertEqual(['five', 'four', 'one', 'two'],
                         sorted(self.reg.keys()))

    def test_values(self):
        self.assertEqual([1, 2, 4, 5], sorted(self.reg.values()))



# This file is part of Ubuntu Continuous Integration configuration framework.
#
# Copyright 2013, 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.


class ConfigError(Exception):
    """Base class for all uci-config exceptions.

    :cvar fmt: A format string that daughter classes override
    """

    fmt = 'Daughter classes should redefine this'

    def __init__(self, **kwargs):
        for key, value in kwargs.items():
            setattr(self, key, value)

    def __unicode__(self):
        return self.fmt.format([], **self.__dict__)

    __repr__ = __unicode__


class ParseError(ConfigError):

    fmt = '{path}({line}): {msg}'

    def __init__(self, path, line, **kwargs):
        super(ParseError, self).__init__(**kwargs)
        self.path = path
        self.line = line


class SectionEmptyName(ParseError):

    fmt = '{path}({line}): Section name cannot be empty.'


class InvalidSyntax(ParseError):

    fmt = '{path}({line}): Not a section nor an option.'


class StoreError(ConfigError):

    def __init__(self, path, **kwargs):
        super(StoreError, self).__init__(**kwargs)
        self.path = path


class InvalidContent(StoreError):

    fmt = '{path}: Content is not utf8.'


class PermissionDenied(StoreError):

    fmt = '{path}: Permission denied.'


class NoSuchFile(StoreError):

    fmt = '{path}: No such file.'


class OptionError(ConfigError):

    def __init__(self, name, **kwargs):
        """Base class for option related errors.

        :param name: The option name.
        """
        super(OptionError, self).__init__(**kwargs)
        self.name = name


class NoSuchConfigOption(OptionError):

    fmt = '{name} does not exist.'


class OptionValueError(OptionError):

    fmt = '{name}: Value "{value}" is not valid.'

    def __init__(self, name, value, **kwargs):
        super(OptionValueError, self).__init__(name, **kwargs)
        self.value = value


class OptionMandatoryValueError(OptionError):

    fmt = '{name} must be set.'


class StackError(ConfigError):
    """Base class for stack related errors."""


class OptionExpansionLoop(StackError):

    fmt = 'Loop involving {refs!r} while expanding "{string}".'

    def __init__(self, string, refs, **kwargs):
        super(OptionExpansionLoop, self).__init__(**kwargs)
        self.string = string
        self.refs = '->'.join(refs)


class ExpandingUnknownOption(StackError):

    fmt = 'Option {name} is not defined while expanding "{string}".'

    def __init__(self, name, string, **kwargs):
        super(ExpandingUnknownOption, self).__init__(**kwargs)
        self.name = name
        self.string = string



# This file is part of Ubuntu Continuous Integration configuration framework.
#
# Copyright 2013 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.


class Registry(object):
    """A class that registers objects to a name.

    There are many places that want to collect related objects and access them
    by a key. This class is designed to allow registering the mapping from key
    to object.
    """

    def __init__(self):
        """Create a new Registry."""
        self.data = {}
        self.help = {}

    def register(self, key, obj, help_string=None):
        """Register a new object to a name.

        :param key: This is the key to use to request the object later.

        :param obj: The object to register.

        :param help_string: Help text for this entry. This may be a string or
                a callable. If it is a callable, it should take two
                parameters (registry, key): this registry and the key that
                the help was registered under.
        """
        if key in self.data:
            raise KeyError('Key {!r} is already registered'.format(key))
        self.data[key] = obj
        self.help[key] = help_string

    def get(self, key):
        """Return the object register()'ed to the given key.

        :param key: The key to obtain the object for. If no object has been
            registered to that key, KeyError will be
            raised.
        :return: The previously registered object.
        """
        return self.data[key]

    def get_help(self, key):
        """Get the help text associated with the given key"""
        the_help = self.help[key]
        if callable(the_help):
            return the_help(self, key)
        return the_help

    def remove(self, key):
        """Remove a registered entry.

        This is mostly for the test suite, but it can be used by others
        """
        del self.data[key]
        del self.help[key]

    def keys(self):
        return sorted(self.data.keys())

    def values(self):
        return self.data.values()



# This file is part of Ubuntu Continuous Integration configuration framework.
#
# Copyright 2013, 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

# A tuple containing the five components of the version number: major, minor,
# micro, releaselevel, and serial. All values except releaselevel are integers;
# the release level is 'dev' or 'final'. The
# version_info value corresponding to the uciconfig version 2.0 is (2, 0, 0,
# 'final', 0).
__version__ = (0, 1, 0, 'final', 0)



# This file is part of Ubuntu Continuous Integration configuration framework.
#
# Copyright 2013, 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import collections
import re


from uciconfig import errors


class Definition(object):

    def __init__(self, pre_comment, post_comment):
        self.pre_comment = pre_comment
        self.post_comment = post_comment

    def __eq__(self, other):
        return (self.pre_comment == other.pre_comment
                and self.post_comment == other.post_comment)


class OptionDefinition(Definition):

    def __init__(self, key, value, pre=None, post=None):
        super(OptionDefinition, self).__init__(pre, post)
        self.key = key
        self.value = value

    def __eq__(self, other):
        return (self.key == other.key
                and self.value == other.value
                and super(OptionDefinition, self).__eq__(other))

    def __repr__(self):
        s = ''
        if self.pre_comment:
            s = s + self.pre_comment
        s = s + self.key + '=' + self.value
        if self.post_comment:
            s = s + self.post_comment
        return s


class SectionDefinition(Definition):

    def __init__(self, name, pre=None, post=None):
        super(SectionDefinition, self).__init__(pre, post)
        self.name = name

    def __eq__(self, other):
        return (self.name == other.name
                and super(SectionDefinition, self).__eq__(other))

    def __repr__(self):
        s = ''
        if self.pre_comment:
            s = s + self.pre_comment
        if self.name is not None:
            s = s + '[' + self.name + ']'
        if self.post_comment:
            s = s + self.post_comment
        return s


class Section(collections.OrderedDict):
    """A section holding options with their associated value and comments.

    This allows serializing configuration files where users can add comments
    and keep them when the configuration file is updated.
    """
    def __init__(self, name, pre=None, post=None):
        super(Section, self).__init__()
        self.name = name
        # The section comments
        self.pre_comment = pre
        self.post_comment = post
        # The option comments. They are handled separatly from the values since
        # relying on the key is enough (renaming options is not a feature ;)
        self.pre = {}
        self.post = {}

    def add(self, key, value, pre=None, post=None):
        """Add a new option.

        :param key: The option name.

        :param value: The option value.

        :param pre: The preceding comment.

        :param post: The comment follwoing the value up to the end of the line
            where it is defined.
        """
        self[key] = value
        self.pre[key] = pre
        self.post[key] = post

    def serialize(self):
        """Serialize the section with the comments.

        :return: The serialized unicode text.
        """
        chunks = []
        if self.name is not None:
            if self.pre_comment is not None:
                chunks.append(self.pre_comment)
            chunks.extend(['[', self.name, ']'])
            if self.post_comment is not None:
                chunks.append(self.post_comment)
            else:
                # Add a newline after each section definition if there is no
                # post comment (which always contain a final newline).
                chunks.append('\n')
        for key, value in self.items():
            pre = self.pre.get(key, None)
            if pre is not None:
                chunks.append(pre)
            chunks.extend([key, ' = ', value])
            post = self.post.get(key, None)
            if post is not None:
                chunks.append(post)
            else:
                # Add a newline after each definition if there is no post
                # comment (which always contain a final newline).
                chunks.append('\n')
        return ''.join(chunks)


class Parser(object):

    comment_re = re.compile(r'(\s*#.*\n)')
    section_re = re.compile(r'\[([^]]*)\]')
    option_re = re.compile(r'\s*([a-zA-Z_][a-zA-Z0-9_.]*)\s*=\s*')

    def __init__(self):
        self.path = None
        self.line = None

    def parse_config(self, path, text):
        """Parse a text containing sections, options and comments.

        :param path: The path where the text is coming from (used for error
            reporting).

        :param text: The unicode string to parse.

        :return: A list of section and option definitions.
        """
        self.path = path
        self.line = 1
        tokens = []
        remaining = text
        pre = None
        while remaining:
            pre_comment = self.comment_re.match(remaining)
            if pre_comment:
                # Buffer the comment lines until a section or an option collect
                # it
                if pre is None:
                    pre = pre_comment.group(1)
                else:
                    pre = pre + pre_comment.group(1)
                remaining = remaining[pre_comment.end():]
                self.line += 1
                continue

            section = self.section_re.match(remaining)
            if section:
                # A new section
                name = section.group(1)
                if name == '':
                    raise errors.SectionEmptyName(self.path, self.line)
                remaining = remaining[section.end():]
                post_comment = self.comment_re.match(remaining)
                if post_comment:
                    post = post_comment.group(1)
                    remaining = remaining[post_comment.end():]
                else:
                    post = None
                tokens.append(SectionDefinition(name, pre, post))
                pre = None
                if post is not None:
                    self.line += 1
                post = None
                continue

            key_matches = self.option_re.match(remaining)
            if key_matches:
                if not tokens:
                    # First option definition without a previous section, it's
                    # the None section.
                    tokens.append(SectionDefinition(None))
                key = key_matches.group(1)
                remaining = remaining[key_matches.end():]
                value, post, remaining = self.parse_value(remaining)
                tokens.append(OptionDefinition(key, value, pre, post))
                if post is not None:
                    self.line += 1
                pre = None
                continue

            if remaining.startswith('\n'):
                # Keep track of the current line
                self.line += 1
                remaining = remaining[1:]
                continue

            if remaining.startswith(' '):
                # Consume spaces to get rid of empty lines
                remaining = remaining[1:]
                continue

            raise errors.InvalidSyntax(self.path, self.line)
        return tokens

    def parse_value(self, text):
        """Parse an option value.

        :param text: The unicode string to parse containing the value.

        :return: A (value, post, remaining) tuple where 'value' is the string
            representing the value, 'post' the associated comment (can be None)
            and 'remaining' the rest of text after parsing.
        """
        value = ''
        post = ''
        cur = 0
        post_last = len(text)
        # ignore leading spaces
        while cur < post_last and text[cur] == ' ':
            cur += 1
        # build value until we encounter end of line or a comment
        while cur < post_last and text[cur] not in ('#', '\n'):
            value += text[cur]
            cur += 1
        # remove trailing spaces
        spaces = ''
        while value.endswith(' '):
            spaces += value[-1]
            value = value[:-1]
        # comment (if present) ends at end of line
        if cur < post_last and text[cur] == '#':
            post = spaces
            while cur < post_last and text[cur] != '\n':
                post += text[cur]
                cur += 1
        if post == '':
            post = None
        # Consume end of line if present
        if cur < post_last and text[cur] == '\n':
            if post is not None:
                post += '\n'
            cur += 1
        remaining = text[cur:]
        return value, post, remaining

    def make_sections(self, tokens):
        """Yields the sections built from the received tokens.

        :param tokens: An iterable of tokens as returned by 'parse_config'.

        :yield: A Section object as soon as one is complete.
        """
        cur = None
        for token in tokens:
            if isinstance(token, SectionDefinition):
                if cur is not None:
                    yield cur
                cur = Section(token.name,
                              token.pre_comment, token.post_comment)
            else:
                cur.add(token.key, token.value,
                        token.pre_comment, token.post_comment)
        if cur is not None:
            yield cur



# This file is part of Ubuntu Continuous Integration configuration framework.
#
# Copyright 2013, 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

import os
import re


from uciconfig import (
    errors,
    options,
    stores,
)


class SectionMatcher(object):
    """Select sections into a given Store.

    This is intended to be used to postpone getting an iterable of sections
    from a store.
    """

    def __init__(self, store):
        self.store = store

    def get_sections(self):
        # This is where we require loading the store so we can see all defined
        # sections.
        sections = self.store.get_sections()
        # Walk the revisions in the order provided
        for store, s in sections:
            if self.match(s):
                yield store, s

    def match(self, section):
        """Does the proposed section match.

        :param section: A Section object.

        :returns: True if the section matches, False otherwise.
        """
        raise NotImplementedError(self.match)


class NameMatcher(SectionMatcher):
    """Select a single section from a store."""

    def __init__(self, store, section_id):
        super(NameMatcher, self).__init__(store)
        self.section_id = section_id

    def match(self, section):
        return section.id == self.section_id


class PathSection(stores.Section):
    """A section providing additional options.

    When the section names are paths, the sections can be selected to provide
    options for a subtree. This section provide automatic options related to
    the full path:

     * relpath: The part of the path that is not in the section name.
     * basename: The last part of the path.
    """

    def __init__(self, section, extra_path):
        super(PathSection, self).__init__(section.id, section.options)
        self.extra_path = extra_path
        self.locals = {'relpath': extra_path,
                       'basename': os.path.basename(extra_path)}

    def get(self, name, default=None, expand=True):
        value = super(PathSection, self).get(name, default)
        if value is not None and expand:
            # expand section local options right now.
            chunks = []
            for is_ref, chunk in iter_option_refs(value):
                if not is_ref:
                    chunks.append(chunk)
                else:
                    ref = chunk[1:-1]
                    if ref in self.locals:
                        chunks.append(self.locals[ref])
                    else:
                        chunks.append(chunk)
            value = ''.join(chunks)
        return value


class StartingPathMatcher(SectionMatcher):
    """A sub path section matcher.

    This selects sections starting with a given path respecting the Store
    order.
    """

    def __init__(self, store, path):
        super(StartingPathMatcher, self).__init__(store)
        self.path = path

    def get_sections(self):
        """Get all sections starting with ``path`` in the store.

        The most generic sections are described first in the store, then more
        specific ones can be provided for reduced scopes.

        The returned sections are therefore returned in the reversed order so
        the most specific ones can be found first.
        """
        path_parts = self.path.rstrip('/').split('/')
        store = self.store
        # Later sections are more specific, they should be returned first
        for _, section in reversed(list(store.get_sections())):
            if section.id is None:
                # The no-name section is always included if present
                yield store, PathSection(section, self.path)
                continue
            section_path = section.id
            if self.path.startswith(section_path):
                section_parts = section_path.rstrip('/').split('/')
                extra_path = '/'.join(path_parts[len(section_parts):])
                yield store, PathSection(section, extra_path)


_option_ref_re = re.compile('({[^{}\n]+})')
"""Describes an expandable option reference.

We want to match the most embedded reference first.

I.e. for '{{foo}}' we will get '{foo}',
for '{bar{baz}}' we will get '{baz}'
"""


def iter_option_refs(string):
    # Split isolated refs so every other chunk is a ref
    is_ref = False
    for chunk in _option_ref_re.split(string):
        yield is_ref, chunk
        is_ref = not is_ref


_shared_stores = {}
_shared_stores_at_exit_installed = False


class Stack(object):
    """A stack of configurations where an option can be defined"""

    def __init__(self, sections_def, store=None, mutable_section_id=None):
        """Creates a stack of sections with an optional store for changes.

        :param sections_def: A list of Section or callables that returns an
            iterable of Section. This defines the Sections for the Stack and
            can be called repeatedly if needed.

        :param store: The optional Store where modifications will be
            recorded. If none is specified, no modifications can be done.

        :param mutable_section_id: The id of the MutableSection where changes
            are recorded. This requires the ``store`` parameter to be
            specified.
        """
        self.sections_def = sections_def
        self.store = store
        self.mutable_section_id = mutable_section_id

    def iter_sections(self):
        """Iterate all the defined sections."""
        # Ensuring lazy loading is achieved by delaying section matching (which
        # implies querying the persistent storage) until it can't be avoided
        # anymore by using callables to describe (possibly empty) section
        # lists.
        for sections in self.sections_def:
            for store, section in sections():
                yield store, section

    def iter_options(self):
        for store, section in self.iter_sections():
            for oname in section.iter_option_names():
                # Only defined options are seen here so they all have a value
                value = section.get(oname, expand=False)
                # The value does not require quoting (for now)
                yield store, section, oname, value

    def get(self, name, expand=True, convert=True):
        """Return the *first* option value found in the sections.

        This is where we guarantee that sections coming from Store are loaded
        lazily: the loading is delayed until we need to either check that an
        option exists or get its value, which in turn may require to discover
        in which sections it can be defined. Both of these (section and option
        existence) require loading the store (even partially).

        :param name: The queried option.

        :param expand: Whether options references should be expanded.

        :param convert: Whether the option value should be converted from
            unicode (do nothing for non-registered options).

        :returns: The value of the option.
        """
        # FIXME: No caching of options nor sections yet -- vila 20110503
        value = None
        found_store = None  # Where the option value has been found
        # If the option is registered, it may provide additional info about
        # value handling
        try:
            opt = options.option_registry.get(name)
        except KeyError:
            # Not registered
            opt = None

        def expand_and_convert(val):
            # This may need to be called in different contexts if the value is
            # None or ends up being None during expansion or conversion.
            if val is not None:
                if expand:
                    val = self._expand_options_in_string(val)
                if opt is None:
                    val = found_store.unquote(val)
                elif convert:
                    val = opt.convert_from_unicode(found_store, val)
            return val

        # First of all, check if the environment can override the configuration
        # value
        if opt is not None and opt.override_from_env:
            value = opt.get_override()
            value = expand_and_convert(value)
        if value is None:
            for store, section in self.iter_sections():
                value = section.get(name)
                if value is not None:
                    found_store = store
                    break
            value = expand_and_convert(value)
            if opt is not None and value is None:
                # If the option is registered, it may provide a default value
                value = opt.get_default()
                value = expand_and_convert(value)
        return value

    def expand_options(self, string, env=None):
        """Expand option references in the string in the configuration context.

        :param string: The string containing option(s) to expand.

        :param env: An option dict defining additional configuration options or
            overriding existing ones.

        :returns: The expanded string.
        """
        return self._expand_options_in_string(string, env)

    def _expand_options_in_string(self, string, env=None, _refs=None):
        """Expand options in the string in the configuration context.

        :param string: The string to be expanded.

        :param env: An option dict defining additional configuration options or
            overriding existing ones.

        :param _refs: Private list (FIFO) containing the options being expanded
            to detect loops.

        :returns: The expanded string.
        """
        if string is None:
            # Not much to expand there
            return None
        if _refs is None:
            # What references are currently resolved (to detect loops)
            _refs = []
        result = string
        # We need to iterate until no more refs appear ({{foo}} will need two
        # iterations for example).
        expanded = True
        while expanded:
            expanded = False
            chunks = []
            for is_ref, chunk in iter_option_refs(result):
                if not is_ref:
                    chunks.append(chunk)
                else:
                    expanded = True
                    name = chunk[1:-1]
                    if name in _refs:
                        raise errors.OptionExpansionLoop(string, _refs)
                    _refs.append(name)
                    value = self._expand_option(name, env, _refs)
                    if value is None:
                        raise errors.ExpandingUnknownOption(name, string)
                    chunks.append(value)
                    _refs.pop()
            result = ''.join(chunks)
        return result

    def _expand_option(self, name, env, _refs):
        if env is not None and name in env:
            # Special case, values provided in env takes precedence over
            # anything else
            value = env[name]
        else:
            value = self.get(name, expand=False, convert=False)
            value = self._expand_options_in_string(value, env, _refs)
        return value

    def _get_mutable_section(self):
        """Get the MutableSection for the Stack.

        This is where we guarantee that the mutable section is lazily loaded:
        this means we won't load the corresponding store before setting a value
        or deleting an option. In practice the store will often be loaded but
        this helps catching some programming errors.
        """
        store = self.store
        section = store.get_mutable_section(self.mutable_section_id)
        return store, section

    def set(self, name, value):
        """Set a new value for the option."""
        store, section = self._get_mutable_section()
        section.set(name, store.quote(value))

    def remove(self, name):
        """Remove an existing option."""
        _, section = self._get_mutable_section()
        try:
            section.remove(name)
        except KeyError:
            raise errors.NoSuchConfigOption(name)

    def __repr__(self):
        # Mostly for debugging use
        return "<config.%s(%s)>" % (self.__class__.__name__, id(self))

    def get_shared_store(self, store):
        """Get a known shared store.

        Store urls uniquely identify them and are used to ensure a single copy
        is shared across all users.

        :param store: The store known to the caller.

        :returns: The store received if it's not a known one, an already known
            otherwise.
        """
        global _shared_stores_at_exit_installed
        the_stores = _shared_stores
        if not _shared_stores_at_exit_installed:
            import atexit

            def save_config_changes():
                for k, store in the_stores.items():
                    store.save_changes()

            atexit.register(save_config_changes)
            _shared_stores_at_exit_installed = True

        url = store.external_url()
        try:
            return the_stores[url]
        except KeyError:
            the_stores[url] = store
            return store


class MemoryStack(Stack):
    """A configuration stack defined from a string.

    This is mainly intended for tests and requires no disk resources.
    """

    def __init__(self, content=None):
        """Create an in-memory stack from a given content.

        It uses a single FileStore without a real file and support reading and
        writing options.

        :param content: The initial content of the store. If None, the store is
            not loaded and ``_load_from_string`` can and should be used if
            needed.
        """
        store = stores.FileStore('<string>')
        if content is not None:
            store._load_from_string(content)
        super(MemoryStack, self).__init__(
            [store.get_sections], store)



#!/usr/bin/env python

# This file is part of Ubuntu Continuous Integration configuration framework.
#
# Copyright 2013, 2014 Canonical Ltd.
#  
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#  
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#  
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.


import setuptools

import uciconfig


setuptools.setup(
    name='uciconfig',
    version='.'.join(str(c) for c in uciconfig.__version__[0:3]),
    description=('Ubuntu Continuous Integration config framework.'),
    author='Vincent Ladeuil',
    author_email='vila+ci@canonical.com',
    url='https://launchpad.net/uci-config',
    license='GPLv3',
    packages=['uciconfig', 'uciconfig.tests'],
)



# This file is part of Ubuntu Continuous Integration configuration framework.
#
# Copyright 2013, 2014 Canonical Ltd.
#
# This program is free software: you can redistribute it and/or modify it under
# the terms of the GNU General Public License version 3, as published by the
# Free Software Foundation.
#
# This program is distributed in the hope that it will be useful, but WITHOUT
# ANY WARRANTY; without even the implied warranties of MERCHANTABILITY,
# SATISFACTORY QUALITY, or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
# General Public License for more details.
#
# You should have received a copy of the GNU General Public License along with
# this program.  If not, see <http://www.gnu.org/licenses/>.

from __future__ import unicode_literals

import unittest


from uciconfig import (
    errors,
    options,
    registries,
    stacks,
    stores,
    tests,
)
from ucitests import (
    assertions,
    fixtures,
)


class TestSectionMatcher(unittest.TestCase):

    def setUp(self):
        super(TestSectionMatcher, self).setUp()
        fixtures.set_uniq_cwd(self)
        self.store = stores.FileStore('foo.conf')
        self.matcher = stacks.NameMatcher

    def test_no_matches_for_empty_stores(self):
        self.store._load_from_string('')
        matcher = self.matcher(self.store, '/bar')
        self.assertEquals([], list(matcher.get_sections()))

    def test_build_doesnt_load_store(self):
        self.matcher(self.store, '/bar')
        self.assertFalse(self.store.is_loaded())

    def get_matching_sections(self, name):
        self.store._load_from_string('''
[foo]
option=foo
[foo/baz]
option=foo/baz
[bar]
option=bar
''')
        matcher = self.matcher(self.store, name)
        return list(matcher.get_sections())

    def test_matching(self):
        sections = self.get_matching_sections('foo')
        assertions.assertLength(self, 1, sections)
        tests.assertSectionContent(self, ('foo', {'option': 'foo'}),
                                   sections[0])

    def test_not_matching(self):
        sections = self.get_matching_sections('baz')
        assertions.assertLength(self, 0, sections)


class TestNameMatcher(unittest.TestCase):

    def setUp(self):
        super(TestNameMatcher, self).setUp()
        self.matcher = stacks.NameMatcher
        # Any simple store is good enough
        fixtures.set_uniq_cwd(self)
        self.store = stores.FileStore('foo.conf')

    def get_matching_sections(self, name):
        self.store._load_from_string('''
[foo]
option=foo
[foo/baz]
option=foo/baz
[bar]
option=bar
''')
        matcher = self.matcher(self.store, name)
        return list(matcher.get_sections())

    def test_matching(self):
        sections = self.get_matching_sections('foo')
        assertions.assertLength(self, 1, sections)
        tests.assertSectionContent(self, ('foo', {'option': 'foo'}),
                                   sections[0])

    def test_not_matching(self):
        sections = self.get_matching_sections('baz')
        assertions.assertLength(self, 0, sections)


class TestPathSection(unittest.TestCase):

    def get_section(self, options, extra_path):
        section = stores.Section('foo', options)
        return stacks.PathSection(section, extra_path)

    def test_simple_option(self):
        section = self.get_section({'foo': 'bar'}, '')
        self.assertEquals('bar', section.get('foo'))

    def test_option_with_extra_path(self):
        section = self.get_section({'foo': 'bar/{basename}'},
                                   'baz')
        self.assertEquals('bar/baz', section.get('foo'))


class TestStartingPathMatcher(unittest.TestCase):

    def setUp(self):
        super(TestStartingPathMatcher, self).setUp()
        fixtures.set_uniq_cwd(self)
        # Any simple store is good enough
        self.store = stores.FileStore('foo.conf')

    def assertSectionIDs(self, expected, path, content):
        self.store._load_from_string(content)
        matcher = stacks.StartingPathMatcher(self.store, path)
        sections = list(matcher.get_sections())
        assertions.assertLength(self, len(expected), sections)
        self.assertEqual(expected, [section.id for _, section in sections])
        return sections

    def test_empty(self):
        self.assertSectionIDs([], '/', '')

    def test_matching_paths(self):
        self.assertSectionIDs(['/foo/bar', '/foo'],
                              '/foo/bar/baz', '''\
[/foo]
[/foo/bar]
''')

    def test_no_name_section_included_when_present(self):
        # Note that other tests will cover the case where the no-name section
        # is empty and as such, not included.
        sections = self.assertSectionIDs(['/foo/bar', '/foo', None],
                                         '/foo/bar/baz', '''\
option = defined so the no-name section exists
[/foo]
[/foo/bar]
''')
        self.assertEquals(['baz', 'bar/baz', '/foo/bar/baz'],
                          [s.locals['relpath'] for _, s in sections])

    def test_order_reversed(self):
        self.assertSectionIDs(['/foo/bar', '/foo'], '/foo/bar/baz', '''\
[/foo]
[/foo/bar]
''')

    def test_unrelated_section_excluded(self):
        self.assertSectionIDs(['/foo/bar', '/foo'], '/foo/bar/baz', '''\
[/foo]
[/foo/qux]
[/foo/bar]
''')

    def test_respect_order(self):
        self.assertSectionIDs(['/foo', '/foo/b', '/foo/bar/baz'],
                              '/foo/bar/baz', '''\
[/foo/bar/baz]
[/foo/qux]
[/foo/b]
[/foo]
''')


class TestBaseStackGet(unittest.TestCase):

    def setUp(self):
        super(TestBaseStackGet, self).setUp()
        fixtures.patch(self, options, 'option_registry',
                       options.OptionRegistry())

    def test_get_first_definition(self):
        store1 = stores.FileStore('<string>')
        store1._load_from_string('foo=bar')
        store2 = stores.FileStore('<string>')
        store2._load_from_string('foo=baz')
        conf = stacks.Stack([store1.get_sections, store2.get_sections])
        self.assertEqual('bar', conf.get('foo'))

    def test_get_with_registered_default_value(self):
        options.option_registry.register(options.Option('foo', 'foo help.',
                                                        default='bar'))
        conf_stack = stacks.Stack([])
        self.assertEqual('bar', conf_stack.get('foo'))

    def test_get_without_registered_default_value(self):
        options.option_registry.register(options.Option('foo', 'foo help.'))
        conf_stack = stacks.Stack([])
        self.assertEqual(None, conf_stack.get('foo'))

    def test_get_without_default_value_for_not_registered(self):
        conf_stack = stacks.Stack([])
        self.assertEqual(None, conf_stack.get('foo'))

    def test_get_for_empty_section_callable(self):
        conf_stack = stacks.Stack([lambda: []])
        self.assertEqual(None, conf_stack.get('foo'))

    def test_get_for_broken_callable(self):
        # Trying to use and invalid callable raises an exception on first use
        conf_stack = stacks.Stack([object])
        self.assertRaises(TypeError, conf_stack.get, 'foo')

    def test_get_registry_option_default_value(self):
        registry = registries.Registry()
        registry.register('bar', 'abc', 'Some help.')
        reg_opt = options.RegistryOption(
            'fooreg', 'A registry option.', registry, default='bar')
        options.option_registry.register(reg_opt)
        conf_stack = stacks.Stack([])
        self.assertEqual('abc', conf_stack.get('fooreg'))


class TestStackWithSimpleStore(unittest.TestCase):

    def setUp(self):
        super(TestStackWithSimpleStore, self).setUp()
        fixtures.patch(self, options, 'option_registry',
                       options.OptionRegistry())
        self.registry = options.option_registry

    def get_conf(self, content=None):
        return stacks.MemoryStack(content)

    def test_override_value_from_env(self):
        fixtures.override_env(self, 'FOO', None)
        self.registry.register(options.Option('foo', 'foo help.', default='b',
                                              override_from_env=['FOO']))
        fixtures.override_env(self, 'FOO', 'quux')
        # Env variable provides a default taking over the option one
        conf = self.get_conf('foo=store')
        self.assertEqual('quux', conf.get('foo'))

    def test_first_override_value_from_env_wins(self):
        fixtures.override_env(self, 'NO_VALUE', None)
        fixtures.override_env(self, 'FOO', None)
        fixtures.override_env(self, 'BAZ', None)
        self.registry.register(options.Option('foo', 'foo help.', default='b',
                                              override_from_env=['NO_VALUE',
                                                                 'FOO',
                                                                 'BAZ']))
        fixtures.override_env(self, 'FOO', 'foo')
        fixtures.override_env(self, 'BAZ', 'baz')
        # The first env var set wins
        conf = self.get_conf('foo=store')
        self.assertEqual('foo', conf.get('foo'))


class TestMemoryStack(unittest.TestCase):

    def test_get(self):
        conf = stacks.MemoryStack('foo=bar')
        self.assertEqual('bar', conf.get('foo'))

    def test_set(self):
        conf = stacks.MemoryStack('foo=bar')
        conf.set('foo', 'baz')
        self.assertEqual('baz', conf.get('foo'))

    def test_no_content(self):
        conf = stacks.MemoryStack()
        # No content means no loading
        self.assertFalse(conf.store.is_loaded())
        # A content can still be provided
        conf.store._load_from_string('foo=bar')
        self.assertEqual('bar', conf.get('foo'))

    # The two following tests are arguably testing the stores but stores do not
    # provide a way to set an option so we need a stack for that.
    def test_set_mark_dirty(self):
        stack = stacks.MemoryStack('')
        assertions.assertLength(self, 0, stack.store.dirty_sections)
        stack.set('foo', 'baz')
        assertions.assertLength(self, 1, stack.store.dirty_sections)
        self.assertTrue(stack.store._needs_saving())

    def test_remove_mark_dirty(self):
        stack = stacks.MemoryStack('foo=bar')
        assertions.assertLength(self, 0, stack.store.dirty_sections)
        stack.remove('foo')
        assertions.assertLength(self, 1, stack.store.dirty_sections)
        self.assertTrue(stack.store._needs_saving())


class TestStackIterSections(unittest.TestCase):

    def test_empty_stack(self):
        conf = stacks.Stack([])
        sections = list(conf.iter_sections())
        assertions.assertLength(self, 0, sections)

    def test_empty_store(self):
        store = stores.FileStore('<string>')
        store._load_from_string('')
        conf = stacks.Stack([store.get_sections])
        sections = list(conf.iter_sections())
        assertions.assertLength(self, 0, sections)

    def test_simple_store(self):
        store = stores.FileStore('<string>')
        store._load_from_string('foo=bar')
        conf = stacks.Stack([store.get_sections])
        tuples = list(conf.iter_sections())
        assertions.assertLength(self, 1, tuples)
        (found_store, found_section) = tuples[0]
        self.assertIs(store, found_store)

    def test_two_stores(self):
        store1 = stores.FileStore('<string>')
        store1._load_from_string('foo=bar')
        store2 = stores.FileStore('<string>')
        store2._load_from_string('bar=qux')
        conf = stacks.Stack([store1.get_sections, store2.get_sections])
        tuples = list(conf.iter_sections())
        assertions.assertLength(self, 2, tuples)
        self.assertIs(store1, tuples[0][0])
        self.assertIs(store2, tuples[1][0])


class TestStackGet(unittest.TestCase):

    def setUp(self):
        super(TestStackGet, self).setUp()
        self.conf = self.get_stack()

    def get_stack(self):
        return stacks.MemoryStack()

    def test_get_for_empty_stack(self):
        self.assertEqual(None, self.conf.get('foo'))


class TestStackGetWithConverter(unittest.TestCase):

    def setUp(self):
        super(TestStackGetWithConverter, self).setUp()
        fixtures.patch(self, options, 'option_registry',
                       options.OptionRegistry())
        self.registry = options.option_registry

    def get_conf(self, content=None):
        return stacks.MemoryStack(content)

    def register_bool_option(self, name, default=None, default_from_env=None):
        b = options.Option(name, 'A boolean.',
                           default=default, default_from_env=default_from_env,
                           from_unicode=options.bool_from_store)
        self.registry.register(b)

    def test_get_default_bool_None(self):
        self.register_bool_option('foo')
        conf = self.get_conf('')
        self.assertEqual(None, conf.get('foo'))

    def test_get_default_bool_True(self):
        self.register_bool_option('foo', 'True')
        conf = self.get_conf('')
        self.assertEqual(True, conf.get('foo'))

    def test_get_default_bool_False(self):
        self.register_bool_option('foo', False)
        conf = self.get_conf('')
        self.assertEqual(False, conf.get('foo'))

    def test_get_default_bool_False_as_string(self):
        self.register_bool_option('foo', 'False')
        conf = self.get_conf('')
        self.assertEqual(False, conf.get('foo'))

    def test_get_default_bool_from_env_converted(self):
        self.register_bool_option('foo', 'True', default_from_env=['FOO'])
        fixtures.override_env(self, 'FOO', 'False')
        conf = self.get_conf('')
        self.assertEqual(False, conf.get('foo'))

    def test_get_default_bool_when_conversion_fails(self):
        self.register_bool_option('foo', default='True')
        conf = self.get_conf('foo=invalid boolean')
        self.assertEqual(True, conf.get('foo'))

    def register_integer_option(self, name,
                                default=None, default_from_env=None):
        i = options.Option(name, 'An integer.',
                           default=default, default_from_env=default_from_env,
                           from_unicode=options.int_from_store)
        self.registry.register(i)

    def test_get_default_integer_None(self):
        self.register_integer_option('foo')
        conf = self.get_conf('')
        self.assertEqual(None, conf.get('foo'))

    def test_get_default_integer(self):
        self.register_integer_option('foo', 42)
        conf = self.get_conf('')
        self.assertEqual(42, conf.get('foo'))

    def test_get_default_integer_as_string(self):
        self.register_integer_option('foo', '42')
        conf = self.get_conf('')
        self.assertEqual(42, conf.get('foo'))

    def test_get_default_integer_from_env(self):
        self.register_integer_option('foo', default_from_env=['FOO'])
        fixtures.override_env(self, 'FOO', '18')
        conf = self.get_conf('')
        self.assertEqual(18, conf.get('foo'))

    def test_get_default_integer_when_conversion_fails(self):
        self.register_integer_option('foo', default='12')
        conf = self.get_conf('foo=invalid integer')
        self.assertEqual(12, conf.get('foo'))

    def register_list_option(self, name, default=None, default_from_env=None):
        l = options.ListOption(name, 'A list.', default=default,
                               default_from_env=default_from_env)
        self.registry.register(l)

    def test_get_default_list_None(self):
        self.register_list_option('foo')
        conf = self.get_conf('')
        self.assertEqual(None, conf.get('foo'))

    def test_get_default_list_empty(self):
        self.register_list_option('foo', '')
        conf = self.get_conf('')
        self.assertEqual([], conf.get('foo'))

    def test_get_default_list_from_env(self):
        self.register_list_option('foo', default_from_env=['FOO'])
        fixtures.override_env(self, 'FOO', '')
        conf = self.get_conf('')
        self.assertEqual([], conf.get('foo'))

    def test_get_with_list_converter_no_item(self):
        self.register_list_option('foo', None)
        conf = self.get_conf('foo=,')
        self.assertEqual([], conf.get('foo'))

    def test_get_with_list_converter_many_items(self):
        self.register_list_option('foo', None)
        conf = self.get_conf('foo=m,o,r,e')
        self.assertEqual(['m', 'o', 'r', 'e'], conf.get('foo'))

    @unittest.skip('Quoting not supported')
    def test_get_with_list_converter_embedded_spaces_many_items(self):
        self.register_list_option('foo', None)
        conf = self.get_conf('foo=" bar", "baz "')
        self.assertEqual([' bar', 'baz '], conf.get('foo'))

    def test_get_with_list_converter_stripped_spaces_many_items(self):
        self.register_list_option('foo', None)
        conf = self.get_conf('foo= bar ,  baz ')
        self.assertEqual(['bar', 'baz'], conf.get('foo'))


class TestStackSet(unittest.TestCase):

    def setUp(self):
        super(TestStackSet, self).setUp()
        fixtures.set_uniq_cwd(self)

    def get_stack(self):
        return stacks.MemoryStack()

    def test_simple_set(self):
        conf = self.get_stack()
        self.assertEqual(None, conf.get('foo'))
        conf.set('foo', 'baz')
        # Did we get it back ?
        self.assertEqual('baz', conf.get('foo'))

    def test_set_creates_a_new_section(self):
        conf = self.get_stack()
        assertions.assertLength(self, 0, list(conf.iter_sections()))
        conf.set('foo', 'baz')
        self.assertEqual, 'baz', conf.get('foo')
        assertions.assertLength(self, 1, list(conf.iter_sections()))


class TestStackRemove(unittest.TestCase):

    def setUp(self):
        super(TestStackRemove, self).setUp()
        fixtures.set_uniq_cwd(self)

    def get_stack(self):
        return stacks.MemoryStack()

    def test_remove_existing(self):
        conf = self.get_stack()
        conf.set('foo', 'bar')
        self.assertEqual('bar', conf.get('foo'))
        conf.remove('foo')
        # Did we get it back ?
        self.assertEqual(None, conf.get('foo'))

    def test_remove_unknown(self):
        conf = self.get_stack()
        with self.assertRaises(errors.NoSuchConfigOption):
            conf.remove('I_do_not_exist')


class TestStackIterMatchingOptions(unittest.TestCase):

    def setUp(self):
        super(TestStackIterMatchingOptions, self).setUp()
        self.conf = stacks.MemoryStack()

    def test_no_options(self):
        self.assertEqual([], list(self.conf.iter_options()))

    def test_options_several_sections(self):
        self.conf.store._load_from_string('''\
one = 1
two = 2
[foo]
one = 1
two = foo
[bar]
one = bar
two = foo
''')
        opts = list(self.conf.iter_options())
        assertions.assertLength(self, 6, opts)
        self.assertEqual([(None, 'one', '1'), (None, 'two', '2'),
                          ('foo', 'one', '1'), ('foo', 'two', 'foo'),
                          ('bar', 'one', 'bar'), ('bar', 'two', 'foo')],
                         [(s.id, n, v) for _, s, n, v in opts])


class TestIterOptionRefs(unittest.TestCase):
    """iter_option_refs is a bit unusual, document some cases."""

    def assertRefs(self, expected, string):
        self.assertEqual(expected, list(stacks.iter_option_refs(string)))

    def test_empty(self):
        self.assertRefs([(False, '')], '')

    def test_no_refs(self):
        self.assertRefs([(False, 'foo bar')], 'foo bar')

    def test_single_ref(self):
        self.assertRefs([(False, ''), (True, '{foo}'), (False, '')], '{foo}')

    def test_broken_ref(self):
        self.assertRefs([(False, '{foo')], '{foo')

    def test_embedded_ref(self):
        self.assertRefs([(False, '{'), (True, '{foo}'), (False, '}')],
                        '{{foo}}')

    def test_two_refs(self):
        self.assertRefs([(False, ''), (True, '{foo}'),
                         (False, ''), (True, '{bar}'),
                         (False, ''), ],
                        '{foo}{bar}')

    def test_newline_in_refs_are_not_matched(self):
        self.assertRefs([(False, '{\nxx}{xx\n}{{\n}}')], '{\nxx}{xx\n}{{\n}}')


class TestStackExpandOptions(unittest.TestCase):

    def setUp(self):
        super(TestStackExpandOptions, self).setUp()
        fixtures.set_uniq_cwd(self)
        fixtures.patch(self, options, 'option_registry',
                       options.OptionRegistry())
        self.registry = options.option_registry
        store = stores.FileStore('foo.conf')
        self.conf = stacks.Stack([store.get_sections], store)

    def assertExpansion(self, expected, string, env=None):
        self.assertEqual(expected, self.conf.expand_options(string, env))

    def test_no_expansion(self):
        self.assertExpansion('foo', 'foo')

    def test_expand_default_value(self):
        self.conf.store._load_from_string('bar=baz')
        self.registry.register(options.Option('foo',  'foo help.',
                                              default='{bar}'))
        self.assertEqual('baz', self.conf.get('foo', expand=True))

    def test_expand_default_from_env(self):
        self.conf.store._load_from_string('bar=baz')
        self.registry.register(options.Option('foo', 'foo help.',
                                              default_from_env=['FOO']))
        fixtures.override_env(self, 'FOO', '{bar}')
        self.assertEqual('baz', self.conf.get('foo', expand=True))

    def test_expand_default_on_failed_conversion(self):
        self.conf.store._load_from_string('baz=bogus\nbar=42\nfoo={baz}')
        self.registry.register(
            options.Option('foo', 'foo help.', default='{bar}',
                           from_unicode=options.int_from_store))
        self.assertEqual(42, self.conf.get('foo', expand=True))

    def test_env_adding_options(self):
        self.assertExpansion('bar', '{foo}', {'foo': 'bar'})

    def test_env_overriding_options(self):
        self.conf.store._load_from_string('foo=baz')
        self.assertExpansion('bar', '{foo}', {'foo': 'bar'})

    def test_simple_ref(self):
        self.conf.store._load_from_string('foo=xxx')
        self.assertExpansion('xxx', '{foo}')

    def test_unknown_ref(self):
        self.assertRaises(errors.ExpandingUnknownOption,
                          self.conf.expand_options, '{foo}')

    def test_indirect_ref(self):
        self.conf.store._load_from_string('''
foo=xxx
bar={foo}
''')
        self.assertExpansion('xxx', '{bar}')

    def test_embedded_ref(self):
        self.conf.store._load_from_string('''
foo=xxx
bar=foo
''')
        self.assertExpansion('xxx', '{{bar}}')

    def test_simple_loop(self):
        self.conf.store._load_from_string('foo={foo}')
        with self.assertRaises(errors.OptionExpansionLoop) as cm:
            self.conf.expand_options('{foo}')
        self.assertEqual('{foo}', cm.exception.string)
        self.assertEqual('foo', cm.exception.refs)
        self.assertEqual('''Loop involving u'foo' while expanding "{foo}".''',
                         unicode(cm.exception))

    def test_indirect_loop(self):
        self.conf.store._load_from_string('''
foo={bar}
bar={baz}
baz={foo}''')
        with self.assertRaises(errors.OptionExpansionLoop) as cm:
            self.conf.expand_options('{foo}')
        self.assertEqual('foo->bar->baz', cm.exception.refs)
        self.assertEqual('{foo}', cm.exception.string)

    def test_list(self):
        self.conf.store._load_from_string('''
foo=start
bar=middle
baz=end
list={foo},{bar},{baz}
''')
        self.registry.register(
            options.ListOption('list', 'A list.'))
        self.assertEqual(['start', 'middle', 'end'],
                         self.conf.get('list', expand=True))

    def test_cascading_list(self):
        self.conf.store._load_from_string('''
foo=start,{bar}
bar=middle,{baz}
baz=end
list={foo}
''')
        self.registry.register(options.ListOption('list', 'A list.'))
        # Register an intermediate option as a list to ensure no conversion
        # happen while expanding. Conversion should only occur for the original
        # option ('list' here).
        self.registry.register(options.ListOption('baz', 'A list.'))
        self.assertEqual(['start', 'middle', 'end'],
                         self.conf.get('list', expand=True))

    def test_pathologically_hidden_list(self):
        self.conf.store._load_from_string('''
foo=bin
bar=go
start={foo
middle=},{
end=bar}
hidden={start}{middle}{end}
''')
        # What matters is what the registration says, the conversion happens
        # only after all expansions have been performed
        self.registry.register(options.ListOption('hidden', 'hidden help.'))
        self.assertEqual(['bin', 'go'],
                         self.conf.get('hidden', expand=True))


class TestStackExpandSectionLocals(unittest.TestCase):
    """Test section local options expansion under various setups."""

    def setUp(self):
        super(TestStackExpandSectionLocals, self).setUp()
        fixtures.set_uniq_cwd(self)

    def get_sections_stack(self, section_path, content):
        """Build a stack from a given content.

        This selects sections with StartingPathMatcher so the section local
        optioins can be expanded.
        """
        store = stores.FileStore('sections.conf')
        store._load_from_string(content)
        store.save()

        class SectionsStack(stacks.Stack):
            """A stack from a store with sections."""

            def __init__(self, store, path):
                super(SectionsStack, self).__init__(
                    [stacks.StartingPathMatcher(store, path).get_sections])

        return SectionsStack(store, section_path)

    def test_expand_locals_empty(self):
        stack = self.get_sections_stack('/home/user/project/', '''
[/home/user/project]
base = {basename}
rel = {relpath}
''')
        self.assertEquals('', stack.get('base', expand=True))
        self.assertEquals('', stack.get('rel', expand=True))

    def test_expand_basename_locally(self):
        stack = self.get_sections_stack('/home/user/project/dir', '''
[/home/user/project]
bfoo = {basename}
''')
        self.assertEquals('dir', stack.get('bfoo', expand=True))

    def test_expand_basename_locally_longer_path(self):
        stack = self.get_sections_stack('/home/user/project/dir/subdir', '''
[/home/user]
bfoo = {basename}
''')
        self.assertEquals('subdir', stack.get('bfoo', expand=True))

    def test_expand_relpath_locally(self):
        stack = self.get_sections_stack('/home/user/project/dir', '''
[/home/user/project]
lfoo = loc-foo/{relpath}
''')
        self.assertEquals('loc-foo/dir', stack.get('lfoo', expand=True))

    def test_locals_dont_leak(self):
        """Ensure we chose the right local across several sections."""
        stack = self.get_sections_stack('/home/user/project/dir', '''
[/home/user]
lfoo = loc-foo/{relpath}
[/home/user/project]
lfoo = loc-foo/{relpath}
''')
        self.assertEquals('loc-foo/dir', stack.get('lfoo', expand=True))
        stack = self.get_sections_stack('/home/user/bar/baz', '''
[/home/user]
lfoo = loc-foo/{relpath}
[/home/user/project]
lfoo = loc-foo/{relpath}
''')
        self.assertEquals('loc-foo/bar/baz', stack.get('lfoo', expand=True))

    def get_2stores_stack(self, section_path,
                          sections_content, no_name_content):
        """Build a stack from two stores with a given content.

        Two stores are built, one with a no-name section, the other with
        several sections.

        The stack will use StartingPathMatcher on the sections store and
        NameMatcher on the no-name one.

        The section local options only exist on the sections store.
        """
        n_store = stores.FileStore('no-name.conf')
        n_store._load_from_string(no_name_content)
        n_store.save()

        s_store = stores.FileStore('sections.conf')
        s_store._load_from_string(sections_content)
        s_store.save()

        class TwoStoresStack(stacks.Stack):
            """A stack from a store with only the no-name section."""

            def __init__(self, s_store, n_store):
                super(TwoStoresStack, self).__init__(
                    [stacks.StartingPathMatcher(s_store,
                                                section_path).get_sections,
                     stacks.NameMatcher(n_store, None).get_sections])

        return TwoStoresStack(s_store, n_store)

    def test_expand_relpath_unknonw_in_global(self):
        stack = self.get_2stores_stack('/home/user/project/dir',
                                       '',
                                       '''
 gfoo = {relpath}
''')
        self.assertRaises(errors.ExpandingUnknownOption,
                          stack.get, 'gfoo', expand=True)

    def test_expand_local_option_locally(self):
        stack = self.get_2stores_stack('/home/user/project/dir',
                                       '''
[/home/user/project]
lfoo = loc-foo/{relpath}
lbar = {gbar}
''',
                                       '''
gfoo = {lfoo}
gbar = glob-bar
''')
        self.assertEquals('glob-bar', stack.get('lbar', expand=True))
        self.assertEquals('loc-foo/dir', stack.get('gfoo', expand=True))


class TestStackCrossSectionsExpand(unittest.TestCase):

    def setUp(self):
        super(TestStackCrossSectionsExpand, self).setUp()
        fixtures.set_uniq_cwd(self)

    def get_config(self, path, string):
        if string is None:
            string = ''

        class PathsStack(stacks.Stack):

            def __init__(self, path):
                store = stores.FileStore('foo.conf')
                super(PathsStack, self).__init__(
                    [stacks.StartingPathMatcher(store, path).get_sections],
                    store)

        c = PathsStack(path)
        c.store._load_from_string(string)
        return c

    def test_dont_cross_unrelated_section(self):
        c = self.get_config('/another/path', '''
[/one/path]
foo = hello
bar = {foo}/2

[/another/path]
bar = {foo}/2
''')
        with self.assertRaises(errors.ExpandingUnknownOption) as cm:
            c.get('bar', expand=True)
        self.assertEqual(
            '''Option foo is not defined while expanding "{foo}/2".''',
            unicode(cm.exception))

    def test_cross_related_sections(self):
        c = self.get_config('/project/branch/path', '''
[/project]
foo = qu

[/project/branch/path]
bar = {foo}ux
''')
        self.assertEqual('quux', c.get('bar', expand=True))


class TestStackCrossStoresExpand(unittest.TestCase):

    def setUp(self):
        super(TestStackCrossStoresExpand, self).setUp()
        fixtures.set_uniq_cwd(self)

    def test_cross_global_locations(self):
        f_store = stores.FileStore('foo.conf')
        f_store._load_from_string('''
[/here]
lfoo = loc-foo
lbar = {gbar}
''')
        f_store.save()
        b_store = stores.FileStore('bar.conf')
        b_store._load_from_string('''
[/]
gfoo = {lfoo}
gbar = glob-bar
''')
        b_store.save()

        class Stack(stacks.Stack):

            def __init__(self, path):
                spm = stacks.StartingPathMatcher
                super(Stack, self).__init__(
                    [spm(f_store, path).get_sections,
                     spm(b_store, path).get_sections])

        stack = Stack('/here')
        self.assertEqual('glob-bar', stack.get('lbar', expand=True))
        self.assertEqual('loc-foo', stack.get('gfoo', expand=True))


class TestSharedStores(unittest.TestCase):

    def test_store_is_shared(self):

        class Stack(stacks.Stack):

            def __init__(self):
                store = self.get_shared_store(stores.FileStore('foo.conf'))
                super(Stack, self).__init__(
                    [store.get_sections],
                    store)

        g1 = Stack()
        g2 = Stack()
        # The two stacks share the same store
        self.assertIs(g1.store, g2.store)
